/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * cairo-compmgr
 * Copyright (C) Nicolas Bruguier 2007 <gandalfn@club-internet.fr>
 * 
 * cairo-compmgr is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * cairo-compmgr is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with cairo-compmgr.  If not, write to:
 * 	The Free Software Foundation, Inc.,
 * 	51 Franklin Street, Fifth Floor
 * 	Boston, MA  02110-1301, USA.
 */
#include <stdio.h>
#include <math.h>

#include "ccm-drawable.h"
#include "ccm-window.h"
#include "ccm-display.h"
#include "ccm-screen.h"
#include "ccm-extension-loader.h"
#include "ccm-magnifier.h"
#include "ccm.h"

enum
{
	CCM_MAGNIFIER_ZOOM_LEVEL,
	CCM_MAGNIFIER_X,
	CCM_MAGNIFIER_Y,
	CCM_MAGNIFIER_HEIGHT,
	CCM_MAGNIFIER_WIDTH,
	CCM_MAGNIFIER_OPTION_N
};

static gchar* CCMMagnifierOptions[CCM_MAGNIFIER_OPTION_N] = {
	"zoom-level",
	"x",
	"y",
	"height",
	"width"
};

static void ccm_magnifier_screen_iface_init(CCMScreenPluginClass* iface);

CCM_DEFINE_PLUGIN (CCMMagnifier, ccm_magnifier, CCM_TYPE_PLUGIN, 
				   CCM_IMPLEMENT_INTERFACE(ccm_magnifier,
										   CCM_TYPE_SCREEN_PLUGIN,
										   ccm_magnifier_screen_iface_init))

struct _CCMMagnifierPrivate
{	
	cairo_surface_t*     surface;
	cairo_t*             ctx;
	CCMRegion*           region;
	CCMConfig*           options[CCM_MAGNIFIER_OPTION_N];
};

#define CCM_MAGNIFIER_GET_PRIVATE(o)  \
   ((CCMMagnifierPrivate*)G_TYPE_INSTANCE_GET_PRIVATE ((o), CCM_TYPE_MAGNIFIER, CCMMagnifierClass))

static void
ccm_magnifier_init (CCMMagnifier *self)
{
	self->priv = CCM_MAGNIFIER_GET_PRIVATE(self);
	self->priv->surface = NULL;
	self->priv->ctx = NULL;
}

static void
ccm_magnifier_finalize (GObject *object)
{
	G_OBJECT_CLASS (ccm_magnifier_parent_class)->finalize (object);
}

static void
ccm_magnifier_class_init (CCMMagnifierClass *klass)
{
	GObjectClass* object_class = G_OBJECT_CLASS (klass);
	
	g_type_class_add_private (klass, sizeof (CCMMagnifierPrivate));

	object_class->finalize = ccm_magnifier_finalize;
}

static void
ccm_magnifier_screen_load_options(CCMScreenPlugin* plugin, CCMScreen* screen)
{
	CCMMagnifier* self = CCM_MAGNIFIER(plugin);
	gint cpt;
	
	for (cpt = 0; cpt < CCM_MAGNIFIER_OPTION_N; cpt++)
	{
		self->priv->options[cpt] = ccm_config_new(screen->number, "magnifier", 
												  CCMMagnifierOptions[cpt]);
	}
	ccm_screen_plugin_load_options(CCM_SCREEN_PLUGIN_PARENT(plugin), screen);
}

static void
ccm_magnifier_set_mag_properties (CCMMagnifier *mag, CCMScreen *self)
{
	g_return_if_fail (mag != NULL);
	g_return_if_fail (self != NULL);

	gint *magrun;
	gfloat *x_scale, *y_scale;

	x_scale = g_object_get_data (G_OBJECT (self), "x_scale");
	y_scale = g_object_get_data (G_OBJECT (self), "y_scale");
	magrun = g_object_get_data (G_OBJECT (self), "mag_run");

	if (!magrun) magrun = g_malloc (sizeof (gint));

	if (!x_scale) x_scale = g_malloc (sizeof (gfloat));
	if (!y_scale) y_scale = g_malloc (sizeof (gfloat));

	*magrun = 1;
	*x_scale = *y_scale = 
		ccm_config_get_float (mag->priv->options [CCM_MAGNIFIER_ZOOM_LEVEL]);

	g_object_set_data (G_OBJECT (self), "x_scale", x_scale);
	g_object_set_data (G_OBJECT (self), "y_scale", y_scale);
	g_object_set_data (G_OBJECT (self), "mag_run", magrun);
}

static void
ccm_magnifier_unset_mag_properties (CCMMagnifier *mag, CCMScreen *self)
{
	g_return_if_fail (mag != NULL);
	g_return_if_fail (self != NULL);

	gint *magrun;
	gfloat *x_scale, *y_scale;

	x_scale = g_object_get_data (G_OBJECT (self), "x_scale");
	y_scale = g_object_get_data (G_OBJECT (self), "y_scale");
	magrun = g_object_get_data (G_OBJECT (self), "mag_run");

	if (!magrun) magrun = g_malloc (sizeof (gint));

	if (!x_scale) x_scale = g_malloc (sizeof (gfloat));
	if (!y_scale) y_scale = g_malloc (sizeof (gfloat));

	*magrun = 1;
	*x_scale = *y_scale = 1;

	g_object_set_data (G_OBJECT (self), "x_scale", x_scale);
	g_object_set_data (G_OBJECT (self), "y_scale", y_scale);
	g_object_set_data (G_OBJECT (self), "mag_run", magrun);
}

static gboolean
ccm_magnifier_screen_paint(CCMScreenPlugin* plugin, CCMScreen* self,
						   cairo_t *ctx)
{
	g_return_val_if_fail(self != NULL, FALSE);

	CCMMagnifier* mag = CCM_MAGNIFIER (plugin);
	static gboolean have_desktop = FALSE;
	gboolean ret = FALSE;
	int x, y, w, h;
	float x_scale, y_scale;

	x = ccm_config_get_integer (mag->priv->options [CCM_MAGNIFIER_X]);
	y = ccm_config_get_integer (mag->priv->options [CCM_MAGNIFIER_Y]);
	w = ccm_config_get_integer (mag->priv->options [CCM_MAGNIFIER_WIDTH]);
	h = ccm_config_get_integer (mag->priv->options [CCM_MAGNIFIER_HEIGHT]);
	x_scale = y_scale = 
		ccm_config_get_float (mag->priv->options [CCM_MAGNIFIER_ZOOM_LEVEL]);

	ccm_magnifier_unset_mag_properties (mag, self);

	ret = ccm_screen_plugin_paint(CCM_SCREEN_PLUGIN_PARENT (plugin), self, ctx);

	ccm_magnifier_set_mag_properties (mag, self);

	if (!mag->priv->surface) {
		cairo_rectangle_t rectangle;
		CCMWindow *cow = ccm_screen_get_overlay_window (self);
		cairo_surface_t* cow_surface =
			ccm_drawable_get_surface (CCM_DRAWABLE (cow));
		if (cow && cow_surface) {
			mag->priv->surface = cairo_surface_create_similar (cow_surface, CAIRO_CONTENT_COLOR, w, h);
			mag->priv->ctx = cairo_create (mag->priv->surface);
			cairo_surface_destroy (mag->priv->surface);
			cairo_rectangle (mag->priv->ctx, 0, 0, w, h);
			cairo_clip (mag->priv->ctx);
			cairo_scale (mag->priv->ctx, x_scale, y_scale);
			cairo_set_source_rgb (mag->priv->ctx, 0, 0, 0);
			cairo_paint(mag->priv->ctx);
			rectangle.x = (double) 0;
			rectangle.y = (double) 0;
			rectangle.width = (double) w;
			rectangle.height = (double) h;
			mag->priv->region = ccm_region_rectangle (&rectangle);
		}
	}

	if (mag->priv->surface && ccm_screen_get_overlay_window (self)) {
		GList *item;

		if (!have_desktop) {
			static gboolean screen_damaged = FALSE;
			
			cairo_set_source_rgb (mag->priv->ctx, 0, 0, 0);
			cairo_paint(mag->priv->ctx);
			if (!screen_damaged) {
				ccm_screen_damage (self);
				screen_damaged = TRUE;
			}
		}
		
		for (item = ccm_screen_get_windows (self); item; item = item->next) {
			CCMWindow* window = item->data;
			
			if (!have_desktop)
				have_desktop = window->is_viewable &&
					ccm_window_get_hint_type (window) == CCM_WINDOW_TYPE_DESKTOP;
			if (!window->is_input_only) {
				CCMRegion *new = ccm_region_copy (mag->priv->region);
				CCMRegion *wreg = ccm_drawable_get_geometry (CCM_DRAWABLE (window));
				if (wreg) {
					ccm_region_intersect (new, wreg);
					if (!ccm_region_empty (new)) {
						if (ccm_window_paint(window, mag->priv->ctx, TRUE))
							ret = TRUE;
					} else {
						ccm_drawable_repair (CCM_DRAWABLE (window));
					}
				}
			}
		}

		if (ret) {
			cairo_set_source_surface (ctx, mag->priv->surface, x, y);
			cairo_paint (ctx);
		}
	}

	return ret;
}

static void
ccm_magnifier_screen_iface_init(CCMScreenPluginClass* iface)
{
	iface->load_options 	= ccm_magnifier_screen_load_options;
	iface->paint 			= ccm_magnifier_screen_paint;
	iface->add_window 		= NULL;
	iface->remove_window 	= NULL;
}

